# encoding: utf-8

import hashlib
import asyncio
import jieba.analyse

from redis import hit


class SimHash(object):


    def __init__(self, path_config=None, num_drawer=4, max_distance=3, num_words=100, retry=10):

        self.path_config  = path_config
        self.num_drawer   = num_drawer
        self.max_distance = max_distance
        self.num_words    = num_words
        # self.num_bits    = 256
        # self.num_bits     = 64
        self.num_bits     = 32
        self.retry        = retry



    def preprocess_keywords(self, doc):

        # pos = ('n' , 's' , 't'  , 'nr' , 'ns' ,
        #        'nt', 'nw', 'nz' , 'v'  , 'vd' ,
        #        'vn', 'a' , 'ad' , 'an' , 'd'  ,
        #         'm', 'q' , 'PER', 'LOC', 'ORG', 'TIME')

        pos = ('n', 'nr', 'ns', 'nt', 'nw', 'nz')
        # pos = ('nr', 'ns', 'nt', 'nw', 'nz')
        # pos = ('ns', 'n', 'vn', 'v')
        # pos = ('ns', 'nt', 'nw', 'nz')

        # kww = jieba.analyse.extract_tags(doc, topK=self.num_words, withWeight=True, allowPOS=pos)
        kww = jieba.analyse.textrank(doc, topK=self.num_words, withWeight=True, allowPOS=pos)

        try:
            keywords, weights = zip(*kww)
        except:
            keywords, weights = [], []

        return keywords, weights


    
    def preprocess_hash(self, keywords):

        # hashkeywords = [format(int(hashlib.sha256(kw.encode()).hexdigest(),16),'b')\
        # hashkeywords = [format(int(hashlib.blake2b(kw.encode(), digest_size=8).hexdigest(),16),'b')\
        hashkeywords = [format(int(hashlib.blake2b(kw.encode(), digest_size=4).hexdigest(),16),'b')\
                        .zfill(self.num_bits)
                        for kw in keywords]

        return hashkeywords



    def preprocess_num(self, num):

        return 1 if 1 == int(num) else -1



    def preprocess_fingerprint(self, hashkeywords, weights):

        length_words = len(hashkeywords)

        fingerprint = ''.join([str(int(sum([self.preprocess_num(hashkeywords[j][i])*weights[j]
                      for j in range(length_words)])>0))
                      for i in range(self.num_bits)])

        return fingerprint



    def preprocess(self, title, content):

        doc               = title + content
        keywords, weights = self.preprocess_keywords(doc)
        hashkeywords      = self.preprocess_hash(keywords)
        fingerprint       = self.preprocess_fingerprint(hashkeywords, weights)

        return fingerprint



    def fit_predict(self, title, content):

        fingerprint = self.preprocess(title, content)

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        # eid, ename  = asyncio.get_event_loop()\
        eid, ename  = loop.run_until_complete(hit(fingerprint,
                                              title,
                                              self.num_drawer,
                                              self.max_distance,
                                              self.path_config,
                                              self.retry))

        return eid, ename



# encoding: utf-8

import os
import time
import random
import aioredis
import traceback

from aioredlock   import Aioredlock
from configparser import ConfigParser



def xor(k1, k2):

    return format(int(k1, 2) ^ int(k2, 2), 'b').count('1')



def keypair(key, seq, num_drawer):

    length = int(len(key) / num_drawer)
    start  = seq * length
    end    = start + length

    key_head = key[start:end]
    key_tail = key[:start] + key[end:]

    return key_head, key_tail



def keyrecover(key_head, key_tail, seq, num_drawer):

    length = int((len(key_head)+len(key_tail)) / num_drawer)
    start  = seq * length
    end    = start + length

    key1   = key_head if 0 == seq else key_tail[:start]
    key2   = '' if 0 == seq or (num_drawer-1) == seq else key_head
    key3   = key_tail if 0 == seq else\
             key_head if (num_drawer-1) == seq else key_tail[len(key1):]

    key    = key1 + key2 + key3

    return key



async def hitting(fingerprint, title, num_drawer, max_distance, redis, lock_manager):

    try:
        async with await lock_manager.lock('lock') as lock:
            assert lock.valid is True

        # check direct
        await redis.select(0)
        event = await redis.lrange(fingerprint, 0, 2)

        if not event:
            found = False

            # check indirect
            for i in range(1, num_drawer+1):
                await redis.select(i)

                # make keypair
                key_head, key_tail = keypair(fingerprint, i-1, num_drawer)

                # candidates
                key_tail_finds = await redis.lrange(key_head, 0, -1)

                if key_tail_finds is None: continue

                # near approximate
                # found
                for key in key_tail_finds:
                    if max_distance >= xor(key_tail, key.decode()):
                        found = True
                        fingerprint_near = keyrecover(key_head, key.decode(), i-1, num_drawer)
                        await redis.select(0)
                        eid, ename = await redis.lrange(fingerprint_near, 0, 2)
                        break

                if found: 
                    break


            # create
            # not found
            if not found:
                eid   = '{0:0>4x}'.format(int(fingerprint, 2)).encode()
                ename = title.encode()
            
            # update "direct"
            await redis.select(0)
            await redis.rpush(fingerprint, eid, ename)

            # update "indirect"
            for i in range(1, num_drawer+1):
                await redis.select(i)
                key_head, key_tail = keypair(fingerprint, i-1, num_drawer)
                await redis.rpush(key_head, key_tail)

        else:
            
            eid, ename = event

        return True, eid.decode(), ename.decode()

    except:

        traceback.print_exc()
        return False, '', ''
        



async def hit(fingerprint, title, num_drawer, max_distance, path_config=None, retry=10):

    """ init
    """
    cfg = ConfigParser()

    path_config = path_config if path_config is not None else\
                  os.path.join(os.path.dirname(__file__), 'config.cfg')
    cfg.read(path_config)
    
    host = cfg.get('redis', 'host')
    port = cfg.getint('redis', 'port')

    conn = await aioredis.create_connection((host, port))
    redis = aioredis.Redis(conn)

    redis_instances = [{'host': host, 'port': port, 'db': num_drawer+1}]   
    lock_manager = Aioredlock(redis_instances)

    """ hit
    """
    count = 0
    while(retry > count):
        flag, eid, ename = await hitting(fingerprint,
                                         title,
                                         num_drawer,
                                         max_distance,
                                         redis,
                                         lock_manager)
        if flag: break

        time.sleep(random.randint(1, 5)) 
        count += 1

    """ close
    """
    redis.close()
    await lock_manager.destroy()

    """ exception
    """
    if retry == count:
        raise ValueError('lock not acquired')

    return eid, ename
